/*********************                                                        */
/*! \file array.h
 ** \verbatim
 ** Top contributors (to current version):
 **   Mathias Preiner, Morgan Deters
 ** This file is part of the CVC4 project.
 ** Copyright (c) 2009-2020 by the authors listed in the file AUTHORS
 ** in the top-level source directory) and their institutional affiliations.
 ** All rights reserved.  See the file COPYING in the top-level source
 ** directory for licensing information.\endverbatim
 **
 ** \brief Array types.
 **
 ** Array types.
 **/

#include <cvc4/cvc4_public.h>

#ifndef CVC4__ARRAY_H
#define CVC4__ARRAY_H

// we get ArrayType right now by #including type.h.
// array.h is still useful for the auto-generated kinds #includes.
#include <cvc4/expr/type.h>

#endif /* CVC4__ARRAY_H */
